#define I2EN 6
#define STA 5
#define STO 4
#define SI 3
#define AA 2

void I2Cinit()
{
	IOCON_PIO0_4=1; // PIO0.4 -> SCL, standard mode
	IOCON_PIO0_5=1; // PIO0.4 -> SDA, standard mode
	I2C0CONSET=(1<<I2EN);	// Enable I2C
	I2C0SCLH=I2C0SCLL=PCLK/100000/2;	// 5us H, 5us L
}

int I2Cframe(unsigned int addr, unsigned char *buf, unsigned int len)
{
	unsigned int s,ret=0;
	//I2C0CONCLR=(1<<SI);	// clear SI
	I2C0CONSET=(1<<STA);	// start
	while ((I2C0CONSET&(1<<SI))==0);	// espera final START
	s=I2C0STAT;
	I2C0CONCLR=(1<<STA);
	if (s==0x8) {
		I2C0DAT=addr;
		I2C0CONCLR=(1<<SI);				// ordena escritura
		while ((I2C0CONSET&(1<<SI))==0);	// espera por byte de dirección
		s=I2C0STAT;
		if (addr&1) {
			// Lectura
			if (s==0x40) {				// ACK de dirección correcto
				I2C0CONSET=(1<<AA);		// contestamos con ACK
				for (;len;len--) {
					if (len<=1) I2C0CONCLR=(1<<AA);	// NACK en último byte
					I2C0CONCLR=(1<<SI);				// ordena lectura
					while ((I2C0CONSET&(1<<SI))==0);	// espera por byte leido
					s=I2C0STAT;
					*buf++=I2C0DAT;
					if (s!=0x50) {					// comprobamos errores
						if (s!=0x58 || len>1) ret=-1;
						break;
					}
				}
			} else ret=-1;	// no ACK en dirección en lectura
		} else {
			// escritura
			if (s==0x18) {	// ACK de dirección correcto
				for (;len;len--) {
					I2C0DAT=*buf++;
					I2C0CONCLR=(1<<SI);				// ordena escritura
					while ((I2C0CONSET&(1<<SI))==0);	// espera por byte escrito
					s=I2C0STAT;
					if (s!=0x28) {ret=-1; break;}	// comprobamos errores
				}
			} else ret=-1;	// no ACK en dirección en escritura
		}
	} else ret=-1;		// error en START
	I2C0CONSET=(1<<STO);	// ordena STOP
	I2C0CONCLR=(1<<SI);	// Libera SCL antes de retornar
	return ret;
}

///////////////////////////////////////////////////////////////////
// Funciones para EEPROMs I2C 
///////////////////////////////////////////////////////////////////

int EEPROM_rd(u32 addr,u8 *pdest, u32 len)
{
	u8 buf[2];

#ifdef EEPROM2BYTES
	buf[0]=addr>>8; buf[1]=addr&0xff;
	if (I2Cframe(0xA0,buf,2)) return -1;
#else
	buf[0]=addr;
	if (I2Cframe(0xA0,buf,1)) return -1;
#endif
	if (I2Cframe(0xA1,pdest,len)) return -1;
	return 0;
}

int EEPROM_wr(u32 addr,u8 *psrc, int len)
{
	u8 buf[EEPROMPAGE+4];
	u32 i,n;

	while (len>0) {
		n=len; if (n>EEPROMPAGE) n=EEPROMPAGE;
#ifdef EEPROM2BYTES
		buf[0]=addr>>8; buf[1]=addr&0xff;
		memcpy(&buf[2],psrc,n);	
#else
		buf[0]=addr;
		memcpy(&buf[1],psrc,n);	
#endif
		for (i=120;i;i--) { // reintentamos mientras NACK
#ifdef EEPROM2BYTES
			if(!I2Cframe(0xA0,buf,n+2)) break;
#else
			if(!I2Cframe(0xA0,buf,n+1)) break;
#endif
		}
		if (!i) return -1;
		addr+=n;
		psrc=&psrc[n];
		len-=n;
	}	
	return 0;
}
