//
 
#include "lpc111x.h"

typedef unsigned char  u8;
typedef unsigned short u16;
typedef unsigned int   u32;

typedef signed char  s8;
typedef signed short s16;
typedef signed int   s32;

#define BAUD0	(115200)
#define PCLK 	(50000000)
#define _delay_us(n) delay_loop((n*(PCLK/1000)-6000)/4000)
#define _delay_ms(n) delay_loop((n*(PCLK/1000)-6)/4)

// Ojo, cuando se ejecuta desde Flash el alineamiento influye
// mucho en la velocidad de ejecución del bucle
void delay_loop(unsigned int) __attribute__ (( naked, aligned(8) ));
void delay_loop(unsigned int d)
{
	asm volatile ("Ldelay_loop: sub r0,#1");
	asm volatile ("  bne Ldelay_loop");
	asm volatile ("  bx lr");
}

// EEPROM I2C 24LC02
#define EEPROMPAGE	(8)		// Size of EEPROM page for writing
#include "i2c.c"

// UART putch

void U0putch(int d)
{
	while(!(U0LSR&0x20));
	U0THR=d;
}

int U0getch()
{
	while(!(U0LSR&1));
	return U0RBR;
}

void U0puts16(unsigned char *p)
{
	while(!(U0LSR&0x20));
	while(*p) U0THR=*p++;
}

//#define putchar(c) U0putch(c)
//#include "printf.c"

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//// 				INTERRUPCIONES
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

#define NLINE	480
#define NCHV	(NLINE/16)
volatile unsigned short linea,oldch;
unsigned short __attribute__((aligned(4))) video[80*NCHV];

#ifdef _RAM_
unsigned char font[32*16]=
#else
unsigned char font[191*16]=
#endif
#include "myfont.h"

void __attribute__((naked,section(".data"))) CT16B0_interrupt(void) 
//void __attribute__((naked)) CT16B0_interrupt(void) 
{
asm volatile(
"	push	{r4,r5}				\n"
"	ldr		r3,=linea			\n"
"	ldrh	r2,[r3]				\n"
"	lsr		r1,r2,#4			\n"
"	cmp		r1,#30				\n"
"	bhs		90f					\n"

"	lsl		r0,r1,#7			\n"
"	lsl		r1,r1,#5			\n"
"	add		r0,r0,r1			\n"
"	ldr		r1,=video			\n"
"	add		r1,r1,r0			\n"	// R1: &video[(linea/16)*160]

"	mov		r0,#0xF				\n"
"	and		r0,r0,r2			\n"
"	ldr		r2,=font			\n"
"	add		r2,r2,r0			\n"	// R2: &font[(linea & 15)]
"	ldr		r3,=GPIO2MASK+(0xFF0<<2)	\n"	// R3: PIO2
"	ldr		r4,=SSP0BAS+8		\n"		// R4: SSP0DR
"	mov		r5,#20				\n"		// R5: contador
"10:							\n"
"	ldrh	r0,[r1]				\n"	// 2 
"	lsl		r0,r0,#4			\n"	// 1
"	strh	r0,[r3]				\n"	// 2
"	lsr		r0,r0,#12			\n"	// 1
"	lsl		r0,r0,#4			\n" // 1
"	ldrb	r0,[r2,r0]			\n" // 2
"	str		r0,[r4]				\n" // 5	
"	nop							\n"	// 1	subtotal: 15 ciclos

"	ldrh	r0,[r1,#2]			\n"	// 2
"	lsl		r0,r0,#4			\n"	// 1
"	strh	r0,[r3]				\n"	// 2
"	lsr		r0,r0,#12			\n"	// 1
"	lsl		r0,r0,#4			\n" // 1
"	ldrb	r0,[r2,r0]			\n" // 2
"	str		r0,[r4]				\n" // 5
"	nop							\n" // 1

"	ldrh	r0,[r1,#4]			\n"	// 2 
"	lsl		r0,r0,#4			\n"	// 1
"	strh	r0,[r3]				\n"	// 2
"	lsr		r0,r0,#12			\n"	// 1
"	lsl		r0,r0,#4			\n" // 1
"	ldrb	r0,[r2,r0]			\n" // 2
"	str		r0,[r4]				\n" // 5
"	nop							\n" // 1

"	ldrh	r0,[r1,#6]			\n"	// 2
"	lsl		r0,r0,#4			\n"	// 1
"	strh	r0,[r3]				\n"	// 2
"	lsr		r0,r0,#12			\n"	// 1
"	lsl		r0,r0,#4			\n" // 1
"	ldrb	r0,[r2,r0]			\n" // 2
"	str		r0,[r4]				\n" // 5

"	add		r1,#8				\n"	// 1

"	sub		r5,#1				\n" // 1
"	bne		10b					\n" // 3	total: 64 ciclos (cada 4 caracteres)

"	nop							\n"
"	mov		r0,#0				\n"
"	strb	r0,[r3]				\n"	// Color en 0 al final de la línea
"	str		r0,[r4]				\n"	// Dato dummy para que se cargue el color

"	ldr		r3,=linea			\n"	// linea++
"	ldrh	r2,[r3]				\n"
"	add		r2,#1				\n"
"	b		95f					\n"

"90:							\n"
"	add		r2,#1				\n"
"	mov		r1,#30				\n"
"	lsl		r1,r1,#4			\n"	// r1=480
"	sub		r1,r2,r1			\n"
"	cmp		r1,#35				\n"
"	beq		92f					\n"
"	cmp		r1,#37				\n"
"	beq		93f					\n"
"	cmp		r1,#49				\n"
"	bls		95f					\n"
"	mov		r2,#0				\n"
"	b		95f					\n"
"92:							\n"	// P0.7 =0
"	ldr		r1,=GPIO0MASK+(1<<(7+2))	\n"
"	mov		r0,#0				\n"
"	str		r0,[r1]				\n"
"	b		95f					\n"
"93:							\n"	// P0.7 =1
"	ldr		r1,=GPIO0MASK+(1<<(7+2))	\n"
"	mov		r0,#1				\n"
"	lsl		r0,r0,#7			\n"
"	str		r0,[r1]				\n"
"95:							\n"
"	strh	r2,[r3]				\n"

"	ldr		r3,=TMR16B0BAS		\n"
"	mov		r0,#4				\n"
"	str		r0,[r3]				\n"	// TMR16B0IR=1<<2;	// Borrado de flag Match MAT2

"	pop		{r4,r5}				\n"
"	bx		lr					\n"
"	.ltorg						\n"
);
}

// Convierte códigos ISO-LATIN a índices de tabla de fuente
unsigned int lat2ix(unsigned int code)
{
	if (code<32) return 0;
	if (code<128) return code-32;
	if (code<128+33) return 0;
	return code-65;
}

// cambia el orden de los bits (8-bits)
int __attribute__((naked)) reversebit8(int d)
{
asm volatile (
"	mov	r1,r0		\n"
"	mov	r0,#0		\n"
"	lsr	r1,r1,#1	\n"
"	adc	r0,r0		\n"
"	lsr	r1,r1,#1	\n"
"	adc	r0,r0		\n"
"	lsr	r1,r1,#1	\n"
"	adc	r0,r0		\n"
"	lsr	r1,r1,#1	\n"
"	adc	r0,r0		\n"
"	lsr	r1,r1,#1	\n"
"	adc	r0,r0		\n"
"	lsr	r1,r1,#1	\n"
"	adc	r0,r0		\n"
"	lsr	r1,r1,#1	\n"
"	adc	r0,r0		\n"
"	lsr	r1,r1,#1	\n"
"	adc	r0,r0		\n"
"	bx	lr			\n"
);
}

#define BLACK	(0)
#define RED		(1)
#define GREEN	(2)
#define BLUE	(4)
#define INTEN	(8)
#define NLCUR	(3125*2)

#define COLOR0	(BLUE|((GREEN+RED+INTEN)<<4))	// Azul de fondo, amarillo intenso de texto

unsigned char color;	// Color de texto (MSB) y de fondo (LSB)
unsigned char posx,posy;	// Posición del cursor

/*
void cls_()				// Usamos 32 bits para acelerar 
{
	unsigned int i,c;
	unsigned int *ps;
	ps=(unsigned int *)video;
	c=color|(color<<16);	// Rellenamos con espacios
	i=NCHV*80/2;
	do {*ps++=c;}while (--i);
	posx=posy=0;
}
*/
void __attribute__((naked)) cls()
{
asm volatile(
"	push	{r4-r5,lr}	\n"
"	ldr		r4,=video	\n"
"	mov		r5,#150		\n"
"	ldr		r0,=color	\n"// EEPROM I2C 24LC02
//#define EEPROMPAGE	(8)		// Size of EEPROM page for writing

"	ldrb	r0,[r0]		\n"
"	lsl		r1,r0,#16	\n"
"	orr		r0,r0,r1	\n"
"	mov		r1,r0		\n"
"	mov		r2,r0		\n"
"	mov		r3,r0		\n"
"1:	stm		r4!,{r0-r3}	\n"
"	stm		r4!,{r0-r3}	\n"
"	sub		r5,r5,#1	\n"
"	bne		1b			\n"
"	ldr		r1,=posx	\n"
"	ldr		r2,=posy	\n"
"	mov		r0,#0		\n"
"	strb	r0,[r1]		\n"
"	strb	r0,[r2]		\n"
"	pop		{r4-r5,pc}	\n"
);
}

 unsigned short __attribute__((naked)) *clsparcial(unsigned short *p, unsigned int n)
{
asm volatile(
"	push	{r4-r5,lr}	\n"
"	ldr		r2,=color	\n"
"	ldrb	r2,[r2]		\n"
"	lsl		r3,r2,#16	\n"
"	orr		r2,r3,r2	\n"
"	mov		r3,r2		\n"
"	mov		r4,r2		\n"
"	mov		r5,r2		\n"
"1:	stm		r0!,{r2-r5}	\n"
"	stm		r0!,{r2-r5}	\n"
"	sub		r1,r1,#1	\n"
"	bne		1b			\n"
"	pop		{r4-r5,pc}	\n"
);
}

/*
void scrool_()	// Usamos 32 bits para acelerar 
{
	unsigned int i,c;
	unsigned int *ps,*pd;
	pd=(unsigned int *)video;
	ps=(unsigned int *)&video[80];
	i=(NCHV-1)*80/2;
	do {*pd++=*ps++;}while (--i);
	c=color|(color<<16);	// Rellenamos con espacios
	i=80/2;
	do {*pd++=c;}while (--i);
}
*/

void __attribute__((naked)) scrool()
{
asm volatile(
"	push	{r4-r6,lr}	\n"
"	ldr		r4,=video	\n"
"	mov		r5,#160		\n"
"	add		r5,r4,r5	\n"
"	mov		r6,#145		\n"
"1:	ldm		r5!,{r0-r3}	\n"
"	stm		r4!,{r0-r3}	\n"
"	ldm		r5!,{r0-r3}	\n"
"	stm		r4!,{r0-r3}	\n"
"	sub		r6,r6,#1	\n"
"	bne		1b			\n"
"	ldr		r0,=color	\n"
"	ldrb	r0,[r0]		\n"
"	lsl		r1,r0,#16	\n"
"	orr		r0,r0,r1	\n"
"	mov		r1,r0		\n"
"	mov		r2,r0		\n"
"	mov		r3,r0		\n"
"	mov		r6,#5		\n"
"2:	stm		r4!,{r0-r3}	\n"
"	stm		r4!,{r0-r3}	\n"
"	sub		r6,r6,#1	\n"
"	bne		2b			\n"
"	pop		{r4-r6,pc}	\n"
"	.ltorg				\n"
);
}

struct {			     		// Parámetros configurables
	unsigned char 	baud;		// Velocidad de la UART
	unsigned char	defcolor;	// Color de texto y de fondo
	unsigned char	flags;		// Controles
} conf;
// Flags
#define	EN_ECHO	(1)
#define	EN_UTF8	(2)
#define EN_TAB8	(4)
#define EN_CR	(8)
#define EN_LF	(0x10)
#define EN_CUR	(0x80)

void vputch(unsigned int ch)
{
	int i,j;
	unsigned short *ps;
	static unsigned char st=0,cx,cy,par[2];
	
	switch(st) {
	case 0:	
		if (ch==27) {st++; break;}
		//////////////// caracteres "normales" ////////////
		switch(ch) {
		case '\n':	
			posx=0;
			if (posy<(NCHV-1)) posy++;
			else scrool();
			break;
		case '\r':
			posx=0;
			break;
		case '\b':
		case 127:
			if (posx) {
				posx--;
				video[posy*80+posx]=color;
			}
			break;
		case '\t':
			if (conf.flags&EN_TAB8) {
				if (posx<72) posx=((posx+8)&0xF8);	// TAB=8
			}else {
				if (posx<76) posx=((posx+4)&0xFC);	// TAB=4
			}
			break;
		default:
			if (posx<80) {
				ch=lat2ix(ch);
				video[posy*80+posx]=color|(ch<<8);
				posx++;
			}
			break;
		}
		break;
	case 1:
		if (ch=='c') AIRCR=0x05FA0004;	// Reset hard
		if (ch!='[') {st=0; break;}
		st++; par[0]=par[1]=0;
		break;
	default:	// recibido <esc>[, ahora toca leer parametros y/o comando
		if (ch==';') {if (st<3) st++; break;}
		if (ch=='?') {par[st-2]='?'; if (st<3) st++; break;}
		if (ch>='0' && ch <='9') {
			par[st-2]*=10;
			par[st-2]+=ch-'0';
			break;
		}
		// Si llegamos aquí es que era un comando
		switch (ch) {
		case 'A':	// Cursor arriba
			if (!par[0]) par[0]=1;
			if (posy>=par[0]) posy-=par[0];
			break;
		case 'B':	// Cursor abajo
			if (!par[0]) par[0]=1;
			if (posy<(NCHV-par[0])) posy+=par[0];
			break;
		case 'C':	// Cursor derecha
			if (!par[0]) par[0]=1;
			if (posx<(80-par[0])) posx+=par[0];
			break;
		case 'D':	// Cursor izquierda
			if (!par[0]) par[0]=1;
			if (posx>=par[0]) posx-=par[0];
			break;
		case 'H':	// Cursor

			if (!par[0]) par[0]++;
			if (!par[1]) par[1]++;
			posx=par[1]-1;
			if (posx>79) posx=79;
			posy=par[0]-1;
			if (posy>NCHV-1) posy=NCHV-1;
			break;
		case 'J':	// Borrar pantalla
			if (par[0]>=2) {cls(); break;}	// Toda
			i=posy*80+posx;
			if (par[0]==1) {	// Desde comienzo
				ps=clsparcial(video,i>>4);	// rápido
				i&=15;
				for(;i;i--) *ps++=color;	// resto
				break;
			}
			j=(i+16)&0xFFFFFFF0;			// Comienzo hasta alineado
			ps=&video[i];
			for (;i<j;i++) *ps++=color;
			clsparcial(ps,(80*NCHV-j)>>4);	// rápido
			break;
		case 'K':	// Borrar línea
			if (par[0]>=2) {		// Toda
				ps=&video[posy*80];
				i=80;
			} else if (par[0]==1) {	// Desde comienzo
				ps=&video[posy*80];
				i=posx;
			} else {// hasta el final
				ps=&video[posy*80+posx];
				i=80-posx;
			}
			do {*ps++=color;} while(--i);
			break;
		case 'm':	// Colores
			if ((!par[0]) && (!par[1])) {color=conf.defcolor; break;}
			ch=2; do {
				i=par[0]; par[0]=par[1];
				if (i==7) color=((color&0xF)<<4)|(color>>4);		// video inverso
				if (i>=30 && i<=37) color = (color&0x0f)|((i-30)<<4);
				if (i>=90 && i<=97) color = (color&0x0f)|((i-90+8)<<4);
				if (i>=40 && i<=47) color = (color&0xf0)|(i-40);
				if (i>=100 && i<=107) color = (color&0xf0)|(i-100+8);
			} while(--ch);
			break;
		case 'l':	// [?25l -> Hide cursor
			if (par[0]=='?' && par[1]==25) conf.flags&=0x7F;
			break;
		case 'h':	// [?25h -> Show cursor
			if (par[0]=='?' && par[1]==25) conf.flags|=0x80;
			break;
		case 's':	cx=posx; cy=posy; break;	// Save cursor pos
		case 'u':	// restore cursor pos
			video[posy*80+posx]=oldch;
			posx=cx; posy=cy;
			//ncur=1; cur=0; 
			break;
		}
		st=0;
		break;
	}
}


void vputs(unsigned char *p)
{
	unsigned short *ps;
	ps=&video[posy*80+posx];
	
	while(*p) *ps++=color|(((*p++)-32)<<8);
}

unsigned int getsc()
{
	unsigned int n;
	while (!(SSP1SR&(1<<2)));
	n=SSP1DR;
	SSP1DR=0xFFFF;
	return reversebit8(n>>2); // Quitamos paridad y stop, y reordenamos bits.
}

/////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//// 				MAIN
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

const unsigned char cursorbm[16*2]={
// Para minúsculas
0x00,    // --------
0x00,    // --------
0x00,    // --------
0x00,    // --------
0x00,    // --------
0x00,    // --------
0xCC,    // ##--##--
0xCC,    // ##--##--
0x30,    // --##----
0x30,    // --##----
0xCC,    // ##--##--
0xCC,    // ##--##--
0x30,    // --##----
0x30,    // --##----
0xCC,    // ##--##--
0xCC,    // ##--##--

// Para mayúsculas
0x33,    // --##--##
0x33,    // --##--##
0xCC,    // ##--##--
0xCC,    // ##--##--
0x33,    // --##--##
0x33,    // --##--##
0xCC,    // ##--##--
0xCC,    // ##--##--
0x33,    // --##--##
0x33,    // --##--##
0xCC,    // ##--##--
0xCC,    // ##--##--
0x33,    // --##--##
0x33,    // --##--##
0xCC,    // ##--##--
0xCC,    // ##--##--

};

int __attribute__ ((noreturn)) main()
{
	int i,n,sc,osc,k,ncur,cur,kmod;
	unsigned char *p;

	static const unsigned char ktbl[]={
		0x00,0x89,0x00,0x85,0x83,0x81,0x82,0x8C,  0x00,0x8A,0x88,0x86,0x84,0x09,0x5C,0x00, // 0x
		0x00,0x00,0x00,0x00,0x00,0x71,0x31,0x00,  0x00,0x00,0x7A,0x73,0x61,0x77,0x32,0x00, // 1x
		0x00,0x63,0x78,0x64,0x65,0x34,0x33,0x00,  0x00,0x20,0x76,0x66,0x74,0x72,0x35,0x00, // 2x
		0x00,0x6E,0x62,0x68,0x67,0x79,0x36,0x00,  0x00,0x00,0x6D,0x6A,0x75,0x37,0x38,0x00, // 3x
		0x00,0x2C,0x6B,0x69,0x6F,0x30,0x39,0x00,  0x00,0x2E,0x2D,0x6C,0xF1,0x70,0x27,0x00, // 4x
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  0x00,0x00,0x0D,0x2B,0x00,0xE7,0x00,0x00, // 5x
		0x00,0x3C,0x00,0x00,0x00,0x00,0x08,0x00,  0x00,0x91,0x00,0x94,0x97,0x00,0x00,0x00, // 6x
		0x90,0x9A,0x92,0x95,0x96,0x98,0x1B,0x00,  0x8B,0x9D,0x93,0x9C,0x9B,0x99,0x00,0x00, // 7x
		0x00,0x00,0x00,0x87																 };// 8x
	static const unsigned char maytab1[]=
		{'=','!','"','#','$','%','&','/','(',')'}; // números en mayúsculas
	static const unsigned char maytab2[]=
		{'<',0x27,'+','-',',','.',0}; // caracteres en minúsculas
	static const unsigned char maytab2b[]=
		{'>','?','*','_',';',':',0}; // caracteres en mayúsculas
	static const unsigned char alttab[]=
		{0x16,0x1E,0x26,0x54,0x5B,0x52,0x5D,0x25,0x4C,0}; // scancode
	static const unsigned char alttabb[]=
		{'|','@','#','[',']','{','}','~','~',0}; //caracteres en ATL GR

	static const unsigned char *speednames[8]={
		"115200","57600","38400","19200",
		"9600","4800","2400","1200"
	};
	static const unsigned short speeddiv[8]={
		(PCLK+115200*8)/(16*115200),
		(PCLK+57600*8)/(16*57600),
		(PCLK+38400*8)/(16*38400),
		(PCLK+19200*8)/(16*19200),
		(PCLK+9600*8)/(16*9600),
		(PCLK+4800*8)/(16*4800),
		(PCLK+2400*8)/(16*2400),
		(PCLK+1200*8)/(16*1200)
	};


	asm volatile ("add sp,#24");	// GCC reserva más espacio del necesario, a pesar de "noreturn"

	// Turn on clock for IOCON 
    SYSAHBCLKCTRL|= BIT16;
	// Release all peripheral resets
	PRESETCTRL=0x0F;	

	// UART. See "Enabling sequence for UART clock" on LPC111x User Manual
	// first assign pin functions for UART	
	IOCON_PIO1_6=1|(2<<3); // RXD, pull-up./bootloader/bt2 -sb 115200 -d /dev/ttyUSB0
	IOCON_PIO1_7=1; // TXD
	// then enable the UART clock
	UARTCLKDIV=1;	// then enable the UART clock
	SYSAHBCLKCTRL= 0x7FFFF;	// Clock for all (also for UART)
	// data baud & format
//   U0LCR=0x83;
//	U0DLM=0;
//	U0DLL=(PCLK+BAUD0*8)/(BAUD0*16);
//	U0LCR=0x3;	// 8-bit 
//	U0IER=1;		// interrupt on RX
//	ISER=1<<21;		// NVIC: Enable UART0 interrupt

	// GPIO
	IOCON_R_PIO0_11=1;	// Por defecto reservado. ¡Serán idiotas en NXP!
	//IOCON_PIO0_7=0;		// GPIO
    GPIO0DIR = (1<<2)|(1<<7)|(1<<8)|(1<<9)|(1<<10)|(1<<11); // Salidas
	GPIO2DIR = 0xFF4;
	GPIO3DIR = 1;
	IOCON_R_PIO1_0=IOCON_R_PIO1_1=IOCON_R_PIO1_2=IOCON_SWDIO_PIO1_3=1;
	GPIO2MASK[1<<2]=-1;
	GPIO3MASK[1]=-1;

	// SSP0
	SSP0CLKDIV=1;			// SPI0 @ 50MHz
	SSP0CPSR=2;				// PCLK/2 
	SSP0CR0=7|(1<<4);		// 8 bit, TI (framed), SCK: 1/1
	SSP0CR1=(1<<1);			// Enable, Master
	IOCON_PIO0_2=1;			// P0.2 como SSEL0
	IOCON_PIO0_9=1;			// P0.9 como MOSI0
	IOCON_SWCLK_PIO0_10=2;	// P0.10 como SCK0
	IOCON_SCK0_LOC=0;		// SCK0 en PIO0.10 

	// SSP1 (PS2)
	SSP1CLKDIV=1;			// SPI1 @ 50MHz
	SSP1CPSR=50;			// PCLK/50 (prescaler) 
	SSP1CR0=10|(1<<6)|(1<<7);		// 11 bit, SPI, POL=1, PHA=1
	SSP1CR1=(1<<1)|(1<<2);	// Enable, Slave
	IOCON_PIO2_0=2; 		// P2.0 como SSEL0
	
	IOCON_PIO2_1=2; 		// P2.1 como SCK
	//IOCON_PIO2_2=2;		// P2.1 como MISO
	IOCON_PIO2_3=2; 		// P2.1 como MOSI

	// Timer
	TMR16B0TCR=3;
	TMR16B0PR=1;		// Prescaler 1/2
	TMR16B0MR0=100; 	// HSYN.width-1
	TMR16B0MR1=799;		// HTOTAl-1
	TMR16B0MR2=115; //125; // Retardo-1 hasta IRQ video (sin latencias en ISR)
	TMR16B0MCR=(2<<3)|(1<<6);	// Reset en MAT1, IRQ en MAT2
	TMR16B0PWMC=1;			// PWM en MAT0
	TMR16B0TCR=1;
	IOCON_PIO0_8=2;		// P0.8 como TMR16B0MAT0

	// Determinismo en interrupciones
	IRQLATENCY=18;
	
	I2Cinit();
	
	////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////
	// Bucle principal
	////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////

	if (i=EEPROM_rd(0,&conf,sizeof(conf))) {
		conf.defcolor=COLOR0;
		conf.baud=0;
		conf.flags=EN_ECHO|EN_UTF8|EN_CR|EN_LF|EN_CUR;
	}

	ISER=1<<16;		// Comienzo de interrupciones de video
	SSP1DR=0xFFFF;	// Precargamos SPI esclavo

reconf:	
	conf.flags|=EN_CUR;	// Cursor ON
	U0LCR=0x83;
	i=speeddiv[conf.baud&7];
	U0DLM=i>>8;
	U0DLL=i;
	U0LCR=0x3;
	U0FCR=7;
	
	color=oldch=conf.defcolor;
	cls();
	memcpy(&font[95*16],cursorbm,16); // Cursor de minúsculas
	kmod=0;	// Sin modificadores en teclado
	ncur=NLCUR; cur=0;
	
	while(1) {
		asm volatile ("WFI");
		if ((--ncur)==0) {	// Parpadeo de cursor
			ncur=NLCUR;
			cur^=1;
			if (cur) oldch=video[posy*80+posx];
			if (posx<80) video[posy*80+posx]=(cur && (conf.flags&0x80))?0x5F00|color : oldch;	
		}
		while (U0LSR&1) {
			n=U0RBR;
			video[posy*80+posx]=oldch ;	// Borrar cursor
			// UTF8?
			if ((conf.flags&EN_UTF8) && (n&0x80)) {
				if ((n>>5)==6) {	// UTF-8 de 2 bytes
					i=U0getch();	// Siguiente byte
					n=(i&0x3F) | ((n&0x1f)<<6);
					if (n>255) continue;
				} else continue;
			}
			vputch(n);					// Imprimir caracter
			ncur=1; cur=0;				// Redibujar cursor de inmediato
		}
		while (SSP1SR&(1<<2)) {	// Scancodes disponibles
			osc=sc;
			sc=getsc();
			if (sc==0xE0) continue;	// Ignorar prefijo
			if (sc==0xF0) {	// Tecla soltada: cambiar modificadores
				while (!(SSP1SR&4));
				osc=sc;
				sc=getsc();
				switch(sc) {
				case 0x12:	kmod&=~(1<<6); break; // Shift Izq
				case 0x59:	kmod&=~(1<<5); break; // Shift Der
				case 0x14:	kmod&=~(1<<4); break; // CTRL Izq/Der
				case 0x11:	kmod&=~(1<<3); break; // ALT / ALT GR
				case 0x58:	kmod&=~(1<<0); break; // Mayúsculas
				}
			}else{	// Tecla pulsada
				switch(sc) {
				case 0x12:	kmod|=(1<<6); break; // Shift Izq
				case 0x59:	kmod|=(1<<5); break; // Shift Der
				case 0x14:	kmod|=(1<<4); break; // CTRL Izq/Der
				case 0x11:	kmod|=(1<<3); break; // ALT / ALT GR
				case 0x58:						 // Mayúsculas	
					if (!(kmod&1)) {
						kmod^=0x80;	// Commuta bit 8 (may)
						memcpy(&font[95*16],&cursorbm[(kmod&0x80)>>3],16); // Cambia cursor
					}
					kmod|=1;
					break;
				case 0x52:
					if(kmod&(1<<3)) goto kmonton;
					if(kmod&0x60) kmod|=1<<11;	// diéresis
					else kmod|=1<<9;	// acento agudo
					break;
				case 0x54:
					if(kmod&(1<<3)) goto kmonton;
					if(kmod&0x60) kmod|=1<<10;	// circumflejo
					else kmod|=1<<8;	// acento grave
					break;
				default:	// Tecla del montón. Procesar
kmonton:			k=ktbl[sc];
					if (k>=0x90 && k<=0x9F) {	// Keypad
						if (osc!=0xE0) {
							if (k<=0x99) {k=k-0x90+'0'; goto sndkey;}
							if (k==0x9A) {k='.'; goto sndkey;}
							if (k==0x9B) {k='*'; goto sndkey;}
							if (k==0x9C) {k='-'; goto sndkey;}
							if (k==0x9D) {k='+'; goto sndkey;}
						} else {	// Teclas edición
							switch (sc) {
							case 0x75:	// UP
								U0puts16("\033[A");
								if (conf.flags&EN_ECHO)	if (posy) {video[posy*80+posx]=oldch ; posy--; ncur=1; cur=0; }
								break;
							case 0x72:	// Down
								U0puts16("\033[B");
								if (conf.flags&EN_ECHO) if (posy<NCHV-1) {video[posy*80+posx]=oldch ; posy++; ncur=1; cur=0; }
								break;
							case 0x6b:	// Left
								U0puts16("\033[D");
								if (conf.flags&EN_ECHO) if (posx) {video[posy*80+posx]=oldch ; posx--; ncur=1; cur=0; }
								break;
							case 0x74:	// right
								U0puts16("\033[C");
								if (conf.flags&EN_ECHO) if (posx<79) {video[posy*80+posx]=oldch ; posx++; ncur=1; cur=0; };
								break;
							case 0x7d:	// Re Pag
								U0puts16("\033[5~");
								break;							
							case 0x7A: 	// Av Pag
								U0puts16("\033[6~");
								break;							
							case 0x6c:	// Home
								U0puts16("\033[H");
								if (conf.flags&EN_ECHO) {video[posy*80+posx]=oldch ; posx=posy=0; ncur=1; cur=0; };
								break;							
							case 0x69:	// End
								U0puts16("\033[F");
								break;							
							case 0x70:	// Insert
								U0puts16("\033[2~");
								break;							
							case 0x71:	// Supr
								U0puts16("\033[3~");
								break;							
							}
						}
						break;
					}
					if (kmod&(1<<4)) {k-=96; goto sndkey;}	// CTRL on
					if (kmod&0xE0) {	// Shift o May
						if ((k>='a' && k<='z') || (k==0xF1) || (k==0xE7)) {k-=32; goto chkacentos;} // Mayúsculas en letras
						if (kmod&0x60) {	// Sólo shift
							if (k>='0' && k<='9') {k=maytab1[k-'0']; goto sndkey;} //Shift en números
							for (i=0;maytab2[i];i++) if (maytab2[i]==k) {k=maytab2b[i];  goto sndkey;}
						}
					}
					if (kmod&(1<<3)) {	// ALT / ALT-GR
						for (i=0;alttab[i];i++) if (alttab[i]==sc) k=alttabb[i];
					}
chkacentos:			if (kmod>>8) { // acentos 
						i=0;					// Grave
						if (kmod&(1<<9)) i=1;	// Agudo
						if (kmod&(1<<10)) i=2;	// Circumflejo
						if (kmod&(1<<11)) i=3;	// Diéresis
						switch (k){
						case 'a':	k=0xE0+i+((i==3)?1:0); break;
						case 'A':	k=0xC0+i+((i==3)?1:0); break;
						case 'e':	k=0xE8+i; break;
						case 'E':	k=0xC8+i; break;
						case 'i':	k=0xEC+i; break;
						case 'I':	k=0xCC+i; break;
						case 'o':	k=0xF2+i+((i==3)?1:0); break;
						case 'O':	k=0xD2+i+((i==3)?1:0); break;
						case 'u':	k=0xF9+i; break;
						case 'U':	k=0xD9+i; break;
						}
					}
sndkey:				if (!k) break;
					if (k=='\r') {
						if (conf.flags&EN_CR) U0putch('\r');
						if (conf.flags&EN_LF) U0putch('\n');
						k='\n'; goto sndecho;
					}
					if (k=='-' && osc==0xE0) k='/';	// tecla KP_/
					
					if ((conf.flags&EN_UTF8) && (k&0x80)) {
						U0putch(0xC0|(k>>6));
						U0putch(0x80|(k&0x3F));
					} else U0putch(k);
sndecho:			if ((conf.flags&EN_ECHO)) {
						video[posy*80+posx]=oldch ;	// Borrar cursor
						vputch(k);
						ncur=1; cur=0;				// Redibujar cursor de inmediato
					}
					kmod&=0xff;	// Fuera acentos
					break;

				case 0x07:	// F12: configuración
					while(1) {
						color=0xF0;
						cls();
						posy=2; posx=4; vputs("Coding: "); 
						posx=16; vputs((conf.flags&EN_UTF8)?"UTF-8":"ISO-LATIN");
						posy=4; posx=4; vputs("Echo: "); 
						posx=16; vputs((conf.flags&EN_ECHO)?"ON":"OFF");
						posy=6; posx=4; vputs("Tab: "); 
						posx=16; vputs((conf.flags&EN_TAB8)?"8":"4");
						
						posy=8; posx=4; vputs("Return: "); 
						i=conf.flags&(EN_CR|EN_LF);
						posx=16;
						if (i==EN_CR) vputs("CR");
						else if(i==EN_LF) vputs("LF");
						else if(i==(EN_CR|EN_LF)) vputs("CR+LF");
						
						posy=10; posx=4; vputs("UART_speed: "); 
						posx=16; vputs(speednames[conf.baud&7]);
						
						posy=12; posx=4; vputs("Defaults");

						posy=14; vputs("Save");

						posx=25; posy=26; vputs("Terminal ANSI 80x30 (VGA/PS2)");
						posx=32; posy=28; vputs("(J. Arias 2021)");
						
						posy=2; posx=40; vputs("Foreground: "); 
						posx=52; color=conf.defcolor&0xF0; vputs("1234ABCDE");
						
						posy=4; posx=40; color=0xF0; vputs("Background: "); 
						posx=52; color=conf.defcolor&0x0F; vputs("         ");
						
						for (i=2;i<16;i+=2) video[i*80+4]|=4;
						for (i=2;i<6;i+=2)  video[i*80+40]|=4;
						
						posy=16; posx=4; color=0xF4; vputs("<escape>");
						
leesc:					osc=sc; sc=getsc();
						switch(sc) {
						case 0xE0:	goto leesc;	// Ignorar
						case 0xF0: 	sc=getsc(); goto leesc;	// soltado
						case 0x21:	conf.flags^=EN_UTF8; break;	// C
						case 0x24:	conf.flags^=EN_ECHO; break;	// E
						case 0x2C:	conf.flags^=EN_TAB8; break;	// T
						case 0x2D:	// R
							i=(conf.flags+EN_CR)&(EN_CR|EN_LF);
							if(!i) i=EN_CR;
							conf.flags=conf.flags&(~(EN_CR|EN_LF))|i;
							break;
						case 0x3C:	conf.baud++; conf.baud&=7; break;	//U
						case 0x2B:	conf.defcolor+=0x10; break;	// F
						case 0x32:	// B
							i=(conf.defcolor+1)&0xF;
							conf.defcolor=(conf.defcolor&0xF0)|i;
							break;
						case 0x23:	// D
							conf.defcolor=COLOR0;
							conf.baud=0;
							conf.flags=EN_ECHO|EN_UTF8|EN_CR|EN_LF|EN_CUR;
							break;
						case 0x1B:	//S
							EEPROM_wr(0,&conf,sizeof(conf));
						case 0x76:	// Escape
							goto reconf;
						}
					}				
					break;
				}
			}
		}	
	}
}
